/* MERG Content-to-template merger
   Command line:
      MERG [switches] template content1 content2 ...

   Valid switches:
      h        allow/Suppress HTML tags in content
      n        Keep/discard name
      l        long file name Acceptable/truncate
      e ERROR  Output error messages to ERROR
               instead of STD_ERR
      q        Quiet mode


mparsec.c Content Parser routines
This file contains the content parser main function call
(ParseContent), as well as internal "helper" functions.
The parser first duplicates the token stream, then creates
a buffer for each read (size = BUFFSIZE). fgets() is not
used to read CR/LF terminated lines; ParseContent calls an
internal function that does a similar job. The output name
is generated by the function ContentName, based on flags
B_LFN and B_NAME.
*/

#include "merg.h"
#include "hash.h"

static char * destname;
static char errstr[BUFFSIZE];

char * ContentName(char * orig, int flags, int count)
{
	char * work = NULL, * format = NULL, * ext = NULL;
	int i, j;

	work = (char *) MemAlloc (sizeof(char) * (strlen(orig) +1));
	work = (char *) memset ((char *)work, '\0', (sizeof(char)*strlen(orig)));
	work = strcpy (work, orig);

	if (!(flags & B_NAME)){
		while(ext = strrchr(work, '.')){
			i = ext - work;
			i = strlen(work) - i;
			for(j=0; j<i; j++)
				ext[j] = '\0';
		}
		
		if ((flags & B_LFN) && (strlen(work) > 8)){
			printf("Cutting name to 8 chars\n");
			for(j=8; j<strlen(work); j++)
				work[j] = '\0';
		}
		sprintf(work, "%s.htm", work);
	}
	else
	{
		if(!(flags & B_LFN))
			format = "out%05i.htm";
		else
			format = "out%i.htm";
			
		sprintf(work, format, count);
	}

	return work;
}

char * getline (FILE *src)
{
	char buff [BUFFSIZE];
	char c, * result;
	int i;

	for(i=0; i < BUFFSIZE; i++)
		buff[i] = 0;
		
	i = 0;

	while ((c=fgetc(src)) != '\n')
		if(c != EOF){
			buff[i] = c;
			i++;
		}
		else{
			return 0;
		}


	result = (char *) MemAlloc (sizeof(char) * (i+1));
	result = strcpy(result, buff);

	return result;
}

int IsSecEnd(char * temp)
{
	if (toupper(temp[0]) == 'E' && toupper(temp[1]) == 'N' && toupper(temp[2]) == 'D')
		return -1;
	else
		return 0;
}

int IsPreDef(char * tag)
{
	if(tag[0] == '_' && tag[1] == '_')          /* All pre-defined sections start */
		return -1;                          /* with 2 underscores (__)        */
	else
		return 0;
}

int IsBlank(char *test)
{
	int i, r = 0;
	
	for (i = 0; i < strlen(test); i++)
		r = r + !isspace(test[i]);

	return !r;
}

int IsValidFileName(char * name){
	int i, r = 0;
	
	for (i = 0; i < strlen(name); i++)
		r = (name[i] != '?' && name[i] != '&' && name[i] != '*') ? r : 1;

	return !r;
}

int MatchDefaults(TOKENSTREAM *dStream, int flags)
{
	int match = 0;
	TOKEN *tmp;

	/* Current implementation matches individual tags;
	   in the future maybe it can use a loop? */


	if(tmp = MatchToken(dStream, "__FILE__")){
		TokenContent(tmp, destname);
		tmp = NULL;
		match++;
	}

	if(tmp = MatchToken(dStream, "__MERG_VER__")){
		TokenContent(tmp, _MERG_VER);
		tmp = NULL;
		match++;
	}

	if(tmp = MatchToken(dStream, "__MERG_BUILD__")){
		TokenContent(tmp, _MERG_BUILD);
		tmp = NULL;
		match++;
	}
	
	if(tmp = MatchToken(dStream, "__MERG_AUTHOR__")){
		TokenContent(tmp, "Jan Michael Iba&ntilde;ez");
		tmp = NULL;
		match++;
	}

	if(tmp = MatchToken(dStream, "__DATE__")){
		TokenContent(tmp, "Sorry, but __DATE__ is currently unimplemented.");
		tmp = NULL;
		match++;
	}
	
	if(tmp = MatchToken(dStream, "__MERG_ERR__")){
		TokenContent(tmp, "Sorry, but __MERG_ERR__ is currently unimplemented.");
		tmp = NULL;
		match++;
	}

	if(tmp = MatchToken(dStream, "__FLAGS__")){
		char * flagstr;
		flagstr = (char *) MemAlloc (sizeof(char) * (46));
		flagstr = (char *) memset ((void *) flagstr, '\0', sizeof(char) * 46);
		if(flags & B_HTML)
			flagstr = strcat (flagstr, "B_HTML  ");
		if(flags & B_QUIET)
			flagstr = strcat (flagstr, "B_QUIET ");
		if(flags & B_LFN)
			flagstr = strcat (flagstr, "B_LFN   ");
		if(flags & B_NAME)
			flagstr = strcat (flagstr, "B_NAME  ");
		if(flags & B_DEBUG)
			flagstr = strcat (flagstr, "B_DEBUG ");

		if(flags = 0)
			flagstr = strcat (flagstr, "No flags");

		TokenContent(tmp, flagstr);
		MemFree(flagstr);
		tmp = NULL;
		match++;
	}

	if(tmp = MatchToken(dStream, "__TEMPLATE__")){
		TokenContent(tmp, dStream->streamname);
		tmp = NULL;
		match++;
	}


	return match;
}

#if defined(_MERG_DUPLICATESTREAM)
int ParseContent(TOKENSTREAM *tStream, char *filename, int flags, FILE *errpipe)
{
	TOKENSTREAM *dStream = DuplicateStream(tStream);
#else
int ParseContent(TOKENSTREAM *dStream, char *filename, int flags, FILE *errpipe)
{
#endif
	FILE * src, * dest;
	FILE * dest2;
	char * dest2name;
	TOKEN * temp;
	char * current, * buff;
	static int count=1;
	int insec=0, match=0, defmatch=0, fileok=0;
	int i;
	int line;
	int buffmon, buffpages;
	long v, minor;
	HASHTABLE * metadata;
	char * lookup;
	
	metadata = createhashtable (75);
	if(!metadata)
		ErrOut ("Warning: Cannot create metadata hash table\n");

	buffmon=0;
	buffpages=1;
	buff = (char *) MemAlloc (BUFFSIZE * buffpages);
	for(i=0; i < (BUFFSIZE * buffpages); i++)
		buff[i]=0;

	if(src = fopen(filename, "r")){
		line = 0;
		while(current = getline(src)){
			line++;
			if(current[0] == '.'){
				if(!fileok){
					fileok++;
					v = 1L;
					minor = 0;
					if(flags & B_DEBUG)
						printf("No version tag found, assuming version 1, minor 0\n");
				}
				if(!(flags & B_QUIET) && (flags & B_DEBUG))
					printf("Checking declarative '%s'...\n", &current[1]);
				if(!IsPreDef(&current[1])){
					if(!insec)
						if (temp = MatchToken(dStream, &current[1])){
							insec=1;
							if(!(flags & B_QUIET) && (flags & B_DEBUG)){
								sprintf(errstr, "Matched section: %s\n", &current[1]);
								ErrOut(errstr);
							}
						}
						else
							insec=-1;
					else if ((insec==1) && IsSecEnd(&current[1])){
						insec=0;
						match++;
						TokenContent(temp, buff);
						if(flags & B_DEBUG)
							printf("%i bytes in buffer stored in %i page(s)\n", strlen(buff), buffpages);
						buffmon=0;
						buffpages=1;
						MemFree(buff);
						buff = (char *) MemAlloc (BUFFSIZE * buffpages);
						for(i=0; i < (BUFFSIZE * buffpages); i++)
							buff[i] = 0;

						if(!(flags & B_QUIET) && (flags & B_DEBUG)){
							printf("Section ended.\n");
						}
						temp = NULL;
					}
					else if ((insec==-1) && IsSecEnd(&current[1])){
						insec=0;
						buffmon=0;
						buffpages=1;
						MemFree(buff);
						buff = (char *) MemAlloc (BUFFSIZE * buffpages);
						for(i=0; i < BUFFSIZE * buffpages; i++)
							buff[i] = 0;

						if(!(flags & B_QUIET) && (flags & B_DEBUG)){
							printf("Section ended.\n");
						}
					}
					else if (insec && temp){
						sprintf(errstr, "Error: unterminated section found, line %i\n", line);
						ErrOut(errstr);
					}
				}
			}
			else if(current[0] == '#' && !insec){
				if(current[1] == '$'){
					if(!fileok){
						fileok++;
						v = atol(&current[2]);
						minor = atol(&current[4]);
						if(flags & B_DEBUG)
							printf("Version is %i, minor %i\n", (int)v, (int)minor);
						if((v > _MERG_VMAJOR) || ((v == _MERG_VMAJOR) && (v > _MERG_VMINOR)))
							ErrOut("Warning: Content needs a higher version of MERG!\n");
						if(v == 0 && minor == 0)
							ErrOut("Warning: Invalid content version, or compatibility mode\n");
					}
					else if(metadata){
						char *name = NULL, *def = NULL;
						int size, off;
						/* Attach to metadata hash table */
						
						for(off = 2; current[off] != '\0' &&  current[off] != ':'; off++)
							;
						
						size = off - 2;
						if (size){
							name = (char *) MemAlloc (sizeof(char) * (size + 1));
							memset ((void *) name, '\0', sizeof(char) * (size + 1));
							name = strncpy (name, &current[2], size);
						}

						off++;
						while(isspace(current[off]) && current[off] != '\0')
							off++;
							

						def = strdup(&current[off]);
						MemFree (name);
						
						if(name && def)
							installname (name, def, metadata);
						
					}
				}
				continue;
			}
			else	if(insec){
				buffmon += (strlen(current) + 1);
				if (buffmon >= ((BUFFSIZE * buffpages)-1)){
					buffpages++;
					buff = (char *) MemRealloc((void *)buff, (BUFFSIZE * buffpages));
				}
				strcat(buff, current);
				strcat(buff, "\n");
			}
			else if(!insec && fileok && !IsBlank(current)){
				sprintf(errstr, "Warning: Stray text in line %i\n", line);
				ErrOut(errstr);
			}
			else if(!fileok){
				break;
			}

			MemFree(current);
		}

		MemFree(buff);

		if(!(flags & B_SIMUL))
			destname = ContentName(filename, flags, count);
		else
			destname = "SIMUL";

		if(flags & B_DEBUG)
			printf ("Testing metadata...\n");

		lookup = getdef("out", metadata);
		if (lookup){
			if(flags & B_DEBUG){
				sprintf(errstr, "User specified name: %s, overriding %s...\n", lookup, destname);
				ErrOut(errstr);
			}
			if(IsValidFileName(lookup))
				destname = strdup(lookup);
			else{
				sprintf(errstr, "Error: invalid filename \"%s\" given by content,\n", lookup);
				ErrOut(errstr);
				sprintf(errstr, "*****: using generated name \"%s\"\n", destname);
				ErrOut(errstr);
			}
		}
		

		if((flags & B_DEBUG) && !(flags & B_SIMUL)){
			sprintf(errstr, "\"%s\" is dest filename\n", destname);
			ErrOut(errstr);
		}
		else if (flags & B_FULLDEBUG)
			ErrOut("Simulation mode, output to STDIO\n");

		/* Match pre-defined (default) sections */
		if (fileok)
			defmatch = MatchDefaults (dStream, flags);
		/* Walk the stream and output */
		if(match || defmatch){
			
			if(flags & B_DEBUG)
				printf("\nOpening dest: %s ...", destname);
			
			if(flags & B_SIMUL)
				dest = stdout;
			else	
				dest = fopen(destname, "w");
				

			if(flags & B_DEBUG)
				printf("okay\n");

			lookup = getdef("also", metadata);
			if(lookup && !(flags & B_SIMUL)){

				if(flags & B_DEBUG){
					sprintf(errstr, "\"%s\" specified as alternate filename\n", lookup);
					ErrOut(errstr);
				}
				if(IsValidFileName(lookup)){
					dest2name = strdup(lookup);
					dest2 = fopen(dest2name, "w");
				}
				else{
					sprintf(errstr, "Error: invalid filename \"%s\" given by content,\n", lookup);
					ErrOut(errstr);
					ErrOut("*****: ignoring alternate filename");
					lookup = NULL;
					dest2name = NULL;
				}
			}

			if(!(flags & B_QUIET))
				printf("%i out of %i matched, %i default section(s).\n", match + defmatch, dStream->namecount, defmatch);
			if (flags & B_DEBUG)
				printf("Writing to file...");
			WalkStream(dStream, dest);
			if (flags & B_DEBUG)
				printf("done.\n");

			if(lookup){
				printf("Alternate filename specified, writing to alternate file...");
				WalkStream(dStream, dest2);
			}

			fclose(dest);
		}
		else{
			sprintf(errstr, "Error: No matches for file \"%s\"\n", filename);
			ErrOut(errstr);
		}

		/* Clean up */

		MemFree(destname);
		fclose(src);

#if defined(_MERG_DUPLICATESTREAM)
		printf("Freeing stream...");
		FreeStream(dStream);
		printf("Okay.\n");
#endif
		count++;
     		return match;
	}
	else if((flags & B_SIMUL) && !src){
		/* Walk the template stream and
		   display each MERG section;
		   replace any "default" sections*/
	}
	else
		return -1;
}
